import re

from django.core.exceptions import ImproperlyConfigured
from django.template import RequestContext
from django.template.loader import render_to_string
from django.template.defaulttags import CsrfTokenNode
from django.utils.datastructures import SortedDict
from django.utils.html import escape
from django.utils.safestring import mark_safe

import django_tables2 as tables

from .utils import BigfootIter
from .utils import render_field
from .utils import convert_camel_case
from .utils import flatatt

__all__ = ('Element', 'Link', 'TemplateElement', 'FormField', 'ElementSet',
    'FormFieldSet', 'Form', 'Table')

class RenderableMixin(object):
    data = None

    def render(self, **kwargs):
        raise ImproperlyConfigured('Subclass must implement render.')

    def __unicode__(self):
        res = self.render()
        return res

    def __str__(self):
        return unicode(self).encode('utf-8')

    def get(self, attr):
        """ Returns the requested attribute or, if the value of the attribute
        is a callable or accessor, the resolved value.
        """

        val = getattr(self, attr)
        if self.data:
            # First, see if we can call resolve()
            try:
                return val.resolve(self.data)
            except (AttributeError, TypeError):
                pass

            # Second, see if it's callable
            try:
                return val(self.data)
            except TypeError:
                try:
                    return val()
                except TypeError:
                    pass

        return val

class Element(RenderableMixin):
    tag = None
    attrs = None
    inner = ""
    __can_be_lazy = ('attrs', 'tag', 'inner', 'html')

    def __init__(self, **kwargs):
        self.attrs = kwargs.get('attrs', {})
        self.tag = kwargs.get('tag')
        self.inner = kwargs.get('inner', self.inner)
        self.html = kwargs.get('html')

    def get_element_template(self, tag):
        """ Renders a string template for the html element from the tag.

        :tag: For example, '<span/>'. <span> would be incorrect as it would not
        support inner HTML or an end tag.
        """

        tag = self.get('tag')
        if tag is None:
            return '%(inner)s'

        tag_check = re.match('^\s*<\s*(?P<tag>[\w:]*)\s*(?P<end_tag>\/?)\s*>\s*$', tag)
        if tag_check is None:
            raise ValueError('tag must be a valid HTML element.')

        tag = tag_check.group('tag')
        has_end_tag = tag_check.group('end_tag') == '/'

        template = '<%s %s>' % (tag, '%(attrs)s')
        if has_end_tag:
            template += '%s</%s>' % ('%(inner)s', tag)

        return template

    def render(self, **kwargs):
        template = self.get_element_template(self.get('tag'))
        return mark_safe(template % {
            'attrs': self.render_attrs(),
            'inner': self.render_inner()
        })

    def render_attrs(self):
        return " ".join(['%s="%s"' % (key, val) for key, val in
            self.get('attrs').items()])

    def render_inner(self):
        html = self.get('html')
        if html:
            # coerce to string in case the html is a bigfoot element
            return unicode(html)
        else:
            return escape(self.get('inner'))

class Link(Element):
    def __init__(self, text, href, attrs=None):
        attrs = attrs or {}
        self.href = href
        super(Link, self).__init__(inner=text, tag='<a/>', attrs=attrs)

    def get(self, attr):
        res = super(Link, self).get(attr)
        if attr == 'attrs':
            res['href'] = self.get('href')
        return res

class Button(Element):
    def __init__(self, **kwargs):
        attrs = kwargs.get('attrs', {})
        attrs.update({'type': 'submit'})
        kwargs['attrs'] = attrs
        kwargs['tag'] = '<input/>'
        super(Button, self).__init__(**kwargs)

class TemplateElement(RenderableMixin):
    template_name = None
    context = None
    request = None
    add_to_context = dict()
    attrs = None

    def __init__(self, template_name=None, context=None, request=None,
    attrs=None):
        context = context or {}
        self.request = request
        self.context = context
        self.template_name = template_name
        self.attrs = attrs or {}

    def render(self, **kwargs):
        context = self.get_context_data(**kwargs)
        return mark_safe(render_to_string(self.get_template_name(), context))

    def get_template_name(self):
        template_name = self.get('template_name')
        if template_name:
            return template_name
        return 'bigfoot/%s.html' % self.__class__.__name__.lower()

    def get_context_data(self, **kwargs):
        context = dict(self.get('context'), **kwargs)

        # Add class attributes to context if anything is listed in
        # add_to_context
        for attr in self.add_to_context:
            val = self.get(attr)
            key = '%s_%s' % (self.__class__.__name__.lower(), attr)
            context[key] = val

        context['attrs'] = mark_safe(flatatt(self.attrs))

        # Return a request context if we have the request
        if self.request:
            return RequestContext(self.request, context)


        return context

class Table(TemplateElement):
    def __init__(self, data, columns, *args, **kwargs):
        self.table_class = kwargs.pop('table_class', tables.Table)
        self.table_attrs = kwargs.pop('table_attrs', {})
        super(Table, self).__init__(*args, **kwargs)
        self.data = data
        self.columns = columns

    def render(self, **kwargs):
        context = self.get_context_data(**kwargs)

        # Create a table renderer
        attrs = SortedDict()
        for col in self.columns.keys():
            attrs[col] = tables.Column(orderable=False)
        Meta = getattr(self.table_class, 'Meta',
            type('Meta', tuple(), {'attrs': {}}))
        Meta.attrs.update(self.table_attrs)
        attrs['Meta'] = Meta
        TableClass = type('BigFootTable', (self.table_class,), attrs)

        data = BigfootIter(self.data, self.columns, context)
        table = TableClass(data)
        context['table'] = table
        return mark_safe(render_to_string(self.get_template_name(), context))

class FormField(TemplateElement):
    add_to_context = ('show_label',)

    def __init__(self, form, field_name, **kwargs):
        self.form = form
        self.name = field_name
        self.show_label = kwargs.pop('show_label', True)
        super(FormField, self).__init__(**kwargs)

    def render(self, **kwargs):
        res = render_field(self.get('name'), self.get('form'), "",
            context = self.get_context_data(**kwargs),
            template = self.get_template_name())
        return res

class FormFields(TemplateElement):
    template_name = 'bigfoot/elementset.html'

    def __init__(self, form, fields, *args, **kwargs):
        self.form = form
        self.fields = fields
        super(FormFields, self).__init__(*args, **kwargs)

    def render(self, **kwargs):
        res = ""
        for field in self.get('fields'):
            res += render_field(self.get('name'), self.get('form'), "",
                context = self.get_context_data(**kwargs),
                template = self.get_template_name())
        return res

class ElementSet(TemplateElement):
    elements = None

    def __init__(self, *elements, **kwargs):
        self.elements = list(elements)
        super(ElementSet, self).__init__(**kwargs)

    def render(self, **kwargs):
        rendered_set = []
        for element in self.elements:
            element.data = self.data or element.data
            rendered_set.append(element.render(**self.context))
        return super(ElementSet, self).render(elements=rendered_set, **kwargs)

    def find(self, **kwargs):
        """ Search for child elements using specified search criterion. The
        following kwargs are supported:

            :element_type: A class of element to search for (e.g., FormField)
        """

        matches = []
        element_type = kwargs.get('element_type')
        if not element_type:
            raise AttributeError('At least one search parameter must be ' +
                'specified.')
        for element in self.elements:
            if isinstance(element, element_type):
                matches.append(element)
            elif isinstance(element, ElementSet):
                matches.extend(element.find(**kwargs))
        return matches

class FormFieldSet(ElementSet):
    def render(self, **kwargs):
        return super(FormFieldSet, self).render(**kwargs)

class Form(FormFieldSet):
    add_to_context = ('method', 'action', 'allow_files')

    def __init__(self, *fields, **kwargs):
        self.method = kwargs.pop('method', 'POST')
        self.action = kwargs.pop('action', '')
        self.allow_files = kwargs.pop('allow_files', False)
        super(Form, self).__init__(*fields, **kwargs)




import logging
import re
import sys

from django.conf import settings
from django.forms.forms import BoundField
from django.template import Context
from django.template.loader import get_template
from django.utils.html import conditional_escape

import elements

def render_field(field, form, form_style, context, template, labelclass=None, layout_object=None, attrs=None):
    """
    Taken from django-crispy-forms. Renders a form field
    
    :param field: Can be a string or a Layout object like `Row`. If it's a layout
        object, we call its render method, otherwise we instantiate a BoundField
        and render it using default template 'uni_form/field.html'
        The field is added to a list that the form holds called `rendered_fields`
        to avoid double rendering fields.

    :param form: The form/formset to which that field belongs to.
    
    :param form_style: A way to pass style name to the CSS framework used.

    :template: Template used for rendering the field.

    :layout_object: If passed, it points to the Layout object that is being rendered.
        We use it to store its bound fields in a list called `layout_object.bound_fields`

    :attrs: Attributes for the field's widget
    """
    FAIL_SILENTLY = False  #getattr(settings, 'CRISPY_FAIL_SILENTLY', True)

    if not hasattr(form, 'rendered_fields'):
        form.rendered_fields = set()

    if isinstance(context, dict):
        context = Context(context)

    if hasattr(field, 'render'):
        return field.render(form, form_style, context)
    else:
        # This allows fields to be unicode strings, always they don't use non ASCII
        try:
            if isinstance(field, unicode):
                field = str(field)
            # If `field` is not unicode then we turn it into a unicode string, otherwise doing
            # str(field) would give no error and the field would not be resolved, causing confusion 
            else:
                field = str(unicode(field))
                
        except (UnicodeEncodeError, UnicodeDecodeError):
            raise Exception("Field '%s' is using forbidden unicode characters" % field)

    try:
        # Injecting HTML attributes into field's widget, Django handles rendering these
        field_instance = form.fields[field]
        if attrs is not None:
            field_instance.widget.attrs.update(attrs)
    except KeyError:
        if not FAIL_SILENTLY:
            raise Exception("Could not resolve form field '%s'." % field)
        else:
            field_instance = None
            logging.warning("Could not resolve form field '%s'." % field, exc_info=sys.exc_info())
            
    if not field in form.rendered_fields:
        form.rendered_fields.add(field)
    else:
        if not FAIL_SILENTLY:
            raise Exception("A field should only be rendered once: %s" % field)
        else:
            logging.warning("A field should only be rendered once: %s" % field, exc_info=sys.exc_info())

    if field_instance is None:
        html = ''
    else:
        bound_field = BoundField(form, field_instance, field)

        template = get_template(template)

        # We save the Layout object's bound fields in the layout object's `bound_fields` list
        if layout_object is not None:
            layout_object.bound_fields.append(bound_field)

        context.update({'field': bound_field, 'labelclass': labelclass, 'flat_attrs': flatatt(attrs or {})})
        html = template.render(context)

    return html


def flatatt(attrs):
    """
    Taken from django.core.utils 
    Convert a dictionary of attributes to a single string.
    The returned string will contain a leading space followed by key="value",
    XML-style pairs.  It is assumed that the keys do not need to be XML-escaped.
    If the passed dictionary is empty, then return an empty string.
    """
    return u''.join([u' %s="%s"' % (k.replace('_', '-'), conditional_escape(v)) for k, v in attrs.items()])

def convert_camel_case(camelstr, delim='_'):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1%s\2' % delim, camelstr)
    return re.sub('([a-z0-9])([A-Z])', r'\1%s\2' % delim, s1).lower()

def formfields(form, *fields, **kwargs):
    """ Shortcut for getting multiple form fields from the same form.

    :type  form: `django.form`
    :param form: The form where the fields are defined.

    :type  fields: `list` of `basestring`
    :param fields: (optional) A list of field names. If not fields are
    specified, all the forms fields are returned.

    :type  wrapper_class: `class`
    :param wrapper_class: (optional) The class to wrap the fields in.
    Defaults to ElementSet.

    :type  field_class: `class`
    :param field_class: (optional) The class that renders the field. Default to
    FormField.

    :returns: An instance of wrapper_class if specified. Otherwise `list`.
    """

    wrapper_class = kwargs.pop('wrapper_class', elements.ElementSet)
    field_class = kwargs.pop('field_class', elements.FormField)
    if not fields:
        fields = form.fields.keys()
    res = [field_class(form, field, **kwargs) for field in fields]
    if wrapper_class:
        res = wrapper_class(*res)
    return res

## {{{ http://code.activestate.com/recipes/496741/ (r1)
class Proxy(object):
    __slots__ = ["_obj", "__weakref__"]
    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)
    
    #
    # proxying (special cases)
    #
    def __getattribute__(self, name):
        return getattr(object.__getattribute__(self, "_obj"), name)
    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_obj"), name)
    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, "_obj"), name, value)
    
    def __nonzero__(self):
        return bool(object.__getattribute__(self, "_obj"))
    def __str__(self):
        return str(object.__getattribute__(self, "_obj"))
    def __repr__(self):
        return repr(object.__getattribute__(self, "_obj"))
    
    #
    # factories
    #
    _special_names = [
        '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__', 
        '__contains__', '__delitem__', '__delslice__', '__div__', '__divmod__', 
        '__eq__', '__float__', '__floordiv__', '__ge__', '__getitem__', 
        '__getslice__', '__gt__', '__hash__', '__hex__', '__iadd__', '__iand__',
        '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__', '__imod__', 
        '__imul__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__', 
        '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__', 
        '__long__', '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__', 
        '__neg__', '__oct__', '__or__', '__pos__', '__pow__', '__radd__', 
        '__rand__', '__rdiv__', '__rdivmod__', '__reduce__', '__reduce_ex__', 
        '__repr__', '__reversed__', '__rfloorfiv__', '__rlshift__', '__rmod__', 
        '__rmul__', '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__', 
        '__rtruediv__', '__rxor__', '__setitem__', '__setslice__', '__sub__', 
        '__truediv__', '__xor__', 'next',
    ]

    _implemented = []
    
    @classmethod
    def _create_class_proxy(cls, theclass):
        """creates a proxy for the given class"""
        
        def make_method(name):
            def method(self, *args, **kw):
                return getattr(object.__getattribute__(self, "_obj"), name)(*args, **kw)
            return method
        
        namespace = {}
        for name in cls._special_names:
            if name in cls._implemented:
                namespace[name] = getattr(cls, name)
            elif hasattr(theclass, name):
                namespace[name] = make_method(name)
        return type("%s(%s)" % (cls.__name__, theclass.__name__), (cls,), namespace)
    
    def __new__(cls, obj, *args, **kwargs):
        """
        creates an proxy instance referencing `obj`. (obj, *args, **kwargs) are
        passed to this class' __init__, so deriving classes can define an 
        __init__ method of their own.
        note: _class_proxy_cache is unique per deriving class (each deriving
        class must hold its own cache)
        """
        try:
            cache = cls.__dict__["_class_proxy_cache"]
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            theclass = cache[obj.__class__]
        except KeyError:
            cache[obj.__class__] = theclass = cls._create_class_proxy(obj.__class__)
        ins = object.__new__(theclass)
        theclass.__init__(ins, obj, *args, **kwargs)
        return ins

class BigfootIter(Proxy):
    """ An iterator that yields a translated version of each data item. The
    translation maps the data item into a dictionary that has the fields
    required by the table being rendered.
    """

    _implemented = ['__iter__']

    def __init__(self, data, translator, context):
        """ Initialization.

        :type  data: Queryset or anything list-like.
        :param data: The data

        :type  translator: `dict`
        :param translator: The translator provides instructions in the form of
        `Accessor`s and renderables (the dict's values) for creating a new
        dictionary with the same keys but with values that can be inserted into
        a template.

        Example:
            >>> BigfootIter(People.object.all(), {
                    'full name':
                    lambda x: '%s %s' % (x.first_name, x.last_name)
                })
        """

        super(BigfootIter, self).__init__(data)
        self.data = data
        self.translator = translator
        self.context = context

    def __iter__(self):
        for row in self.data:
            vals = dict()
            for name, trans in self.translator.items():
                rendered = False
                try:
                    # Try as an accessor
                    val = trans.resolve(row)
                    rendered = True
                except (TypeError, AttributeError):
                    pass

                if not rendered:
                    # Try as a renderable
                    trans.data = row
                    val = trans.render(**self.context)

                vals[name] = val

            yield vals




import sys

from setuptools import setup, find_packages

install_requires = [
    'django >= 1.4',
    'django-tables2 == 0.13'
]

setup(
    name = "django-bigfoot",
    version = '0.1',
    description = "A Django library for defining HTML building blocks (e.g.," +
        " forms, form fields, tables) that can be easily rendered into HTML.",
    url = "https://github.com/setaris/django-bigfoot",
    author = "Setaris",
    author_email = "support@setaris.com",
    packages = find_packages(),
    install_requires = install_requires,
)
